require 'json'
require 'zlib'
require 'net/http'

module ActiveRecord
  class QueryCounter
    class << self
      attr_accessor :queries
    end

    IGNORED_SQL = [
      /^PRAGMA (?!(table_info))/,
      /^SELECT currval/,
      /^SELECT CAST/,
      /^SELECT @@IDENTITY/,
      /^SELECT @@ROWCOUNT/,
      /^SAVEPOINT/,
      /FROM pg_attribute/,
      /JOIN pg_attribute/,
      /JOIN pg_namespace/,
      /SHOW search_path/,
      /^BEGIN$/,
      /^ROLLBACK$/,
      /^ROLLBACK TO SAVEPOINT/,
      /^RELEASE SAVEPOINT/,
      /^SHOW max_identifier_length/
    ]

    def call(name, start, finish, message_id, values)
      return if 'CACHE' == values[:name]
      return if IGNORED_SQL.any? { |r| values[:sql] =~ r }

      self.class.queries ||= []
      self.class.queries << {
        duration: finish - start,
        sql: values[:sql]
      }
    end
  end
end

module TestScore
  def self.rails?
    defined?(ActiveRecord) && defined?(ActiveSupport) && defined?(ActiveSupport::Notifications)
  end

  def self.queries
    return [] unless rails?

    ActiveRecord::QueryCounter.queries
  end

  def self.log_suite_started
    @start_time = Time.now

    if rails?
      ActiveSupport::Notifications.subscribe('sql.active_record', ActiveRecord::QueryCounter.new)
    end
  end

  def self.spec_started(example)
    @results ||= {}
    @results[example.id] = { started_at: Time.now }

    if rails?
      ActiveRecord::QueryCounter.queries = []
    end
  end

  def self.spec_ended(example)
    @results[example.id] = {
      duration: Time.now - @results[example.id][:started_at],
      description: example.description,
      metadata: example.metadata.slice(:line_number, :file_path),
      queries: queries
    }
  end

  def self.log_results
    puts "\n\nReporting test suite performance to TestScore."
    host = ENV.fetch('TESTSCORE_REPORT_HOST', 'https://testscore.io')
    uri = URI("#{host}/results/#{@api_key}")
    req = Net::HTTP::Post.new(
      uri,
      'Content-Type' => 'application/json',
      'Content-Encoding' => 'gzip',
    )
    gzip = Zlib::GzipWriter.new(StringIO.new)
    gzip << { examples: @results }.to_json
    req.body = gzip.close.string
    http = Net::HTTP.new(uri.host, uri.port)
    http.use_ssl = (uri.scheme == "https")
    response = http.request(req).body

    if response && response.length > 0
      jsonResponse = JSON.parse(response)

      if jsonResponse['error']
        puts "Unable to report results to TestScore: #{jsonResponse['error']}"
      end
    end
  rescue => e
    puts "Unable to report results to TestScore: #{e.message}"
  end

  def self.start(api_key)
    @api_key = api_key

    RSpec.configure do |config|
      config.before(:each) do |example|
        TestScore.spec_started(example)
      end

      config.after(:each) do |example|
        TestScore.spec_ended(example)
      end

      config.before(:suite) do |suite|
        TestScore.log_suite_started
      end

      config.after(:suite) do |suite|
        TestScore.log_results
      end
    end
  end
end
